import spacy_curated_transformers.cli.debug_pieces
from spacy.cli import app
from typer.testing import CliRunner


def test_debug_pieces():
    result = CliRunner().invoke(app, ["debug", "pieces", "--help"])
    assert result.exit_code == 0



from typing import Any, Callable, Iterable, List, Optional, Tuple

from spacy import Errors as SpacyErrors
from spacy.tokens import Doc

from thinc.api import Model, deserialize_attr, serialize_attr
from thinc.types import Floats2d, Ragged

from ..errors import Errors
from .output import TransformerModelOutput
from .pooling import with_ragged_last_layer, with_ragged_layers
from .types import (
    PoolingModelT,
    ScalarWeightModelT,
    ScalarWeightOutT,
    TransformerListenerModelT,
    TransformerModelT,
    WithRaggedLastLayerModelT,
    WithRaggedLayersModelT,
    WrappedTransformerAndListenerBackpropT,
    WrappedTransformerAndListenerInT,
    WrappedTransformerAndListenerModelT,
    WrappedTransformerAndListenerOutT,
)


def build_transformer_layers_listener_v1(
    layers: int,
    width: int,
    pooling: PoolingModelT,
    upstream: str = "*",
    grad_factor: float = 1.0,
) -> TransformerListenerModelT:
    """Construct a listener layer that communicates with one or more upstream Transformer
    components. This layer extracts the output of all transformer layers and performs
    pooling over the individual pieces of each Doc token, returning their corresponding
    representations.

    layers (int):
        The the number of layers produced by the upstream transformer component,
        excluding the embedding layer.
    width (int):
        The width of the vectors produced by the upstream transformer component.
    pooling (PoolingModelT):
        Model that is used to perform pooling over the piece representations.
    upstream (str):
        A string to identify the 'upstream' Transformer component
        to communicate with. The upstream name should either be the wildcard
        string '*', or the name of the Transformer component.

        In almost all cases, the wildcard string will suffice as there'll only be one
        upstream Transformer component. But in certain situations, e.g: you have disjoint
        datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
        downstream task requires its own token representations, you could end up with
        more than one Transformer component in the pipeline.
    grad_factor (float):
        Factor to multiply gradients with.
    """
    model: TransformerListenerModelT = Model(
        name="transformer_layers_listener",
        forward=transformer_layers_listener_forward,
        dims={"nO": width},
        layers=[with_ragged_layers(pooling)],
        attrs={
            "grad_factor": grad_factor,
            "layers": layers + 1,
        },
        refs={"pooling": pooling},
    )
    ListenerStateUtils.init_state(
        model, upstream_name=upstream, requires_all_layer_outputs=True
    )
    return model


def build_last_transformer_layer_listener_v1(
    width: int,
    pooling: PoolingModelT,
    upstream: str = "*",
    grad_factor: float = 1.0,
) -> TransformerListenerModelT:
    """Construct a listener layer that communicates with one or more upstream Transformer
    components. This layer extracts the output of the last transformer layer and performs
    pooling over the individual pieces of each Doc token, returning their corresponding
    representations.

    width (int):
        The width of the vectors produced by the upstream transformer component.
    pooling (Model):
        Model that is used to perform pooling over the piece representations.
    upstream (str):
        A string to identify the 'upstream' Transformer component
        to communicate with. The upstream name should either be the wildcard
        string '*', or the name of the Transformer component.

        In almost all cases, the wildcard string will suffice as there'll only be one
        upstream Transformer component. But in certain situations, e.g: you have disjoint
        datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
        downstream task requires its own token representations, you could end up with
        more than one Transformer component in the pipeline.
    grad_factor (float):
        Factor to multiply gradients with.
    """
    model: TransformerListenerModelT = Model(
        name="last_transformer_layer_listener",
        forward=last_transformer_layer_listener_forward,
        dims={"nO": width},
        layers=[with_ragged_last_layer(pooling)],
        attrs={"grad_factor": grad_factor},
        refs={"pooling": pooling},
    )
    ListenerStateUtils.init_state(
        model, upstream_name=upstream, requires_all_layer_outputs=False
    )
    return model


def build_scalar_weighting_listener_v1(
    width: int,
    weighting: ScalarWeightModelT,
    pooling: PoolingModelT,
    upstream: str = "*",
    grad_factor: float = 1.0,
) -> TransformerListenerModelT:
    """Construct a listener layer that communicates with one or more upstream Transformer
    components. This layer calculates a weighted representation of all transformer layer
    outputs and performs pooling over the individual pieces of each Doc token, returning
    their corresponding representations.

    Requires its upstream Transformer components to return all layer outputs from
    their models.

    width (int):
        The width of the vectors produced by the upstream transformer component.
    weighting (Model):
        Model that is used to perform the weighting of the different layer outputs.
    pooling (Model):
        Model that is used to perform pooling over the piece representations.
    upstream (str):
        A string to identify the 'upstream' Transformer component
        to communicate with. The upstream name should either be the wildcard
        string '*', or the name of the Transformer component.

        In almost all cases, the wildcard string will suffice as there'll only be one
        upstream Transformer component. But in certain situations, e.g: you have disjoint
        datasets for certain tasks, or you'd like to use a pre-trained pipeline but a
        downstream task requires its own token representations, you could end up with
        more than one Transformer component in the pipeline.
    grad_factor (float):
        Factor to multiply gradients with.
    """
    model: TransformerListenerModelT = Model(
        name="scalar_weighting_listener",
        forward=scalar_weighting_listener_forward,
        dims={"nO": width},
        layers=[weighting, with_ragged_last_layer(pooling)],
        attrs={
            "grad_factor": grad_factor,
        },
    )
    ListenerStateUtils.init_state(
        model, upstream_name=upstream, requires_all_layer_outputs=True
    )
    return model


# We need to store the listener state in the attributes dict of the listener model.
# This custom class is used to ensure that this state is not serialized to disk.
class _ListenerNonPersistentState:
    batch_id: Optional[int] = None
    outputs: Optional[TransformerModelOutput] = None
    backprop: Optional[Callable[[List[List[Ragged]], Tuple[int]], Any]] = None


@serialize_attr.register(_ListenerNonPersistentState)
def serialize_listener_non_persistent_state(
    _, value: _ListenerNonPersistentState, name: str, model
) -> bytes:
    return bytes()


@deserialize_attr.register(_ListenerNonPersistentState)
def deserialize_listener_non_persistent_state(
    _, value: bytes, name: str, model
) -> _ListenerNonPersistentState:
    return _ListenerNonPersistentState()


class ListenerStateUtils:
    """This class provides helper functions to manage the internal state of
    `Model`s that act as listeners of the `Transformer` pipe."""

    SENTINEL = "_TRANSFORMER_LISTENER"
    USE_DOC_ANNOTATIONS_FOR_PREDICTION = "_USE_DOC_ANNOTATIONS_FOR_PREDICTION"
    REQUIRES_ALL_LAYER_OUTPUTS = "_REQUIRES_ALL_LAYER_OUTPUTS"

    @classmethod
    def is_listener(cls, model: Model) -> bool:
        return cls.SENTINEL in model.attrs

    @classmethod
    def init_state(
        cls,
        listener: TransformerListenerModelT,
        *,
        upstream_name: str = "*",
        requires_all_layer_outputs: bool
    ):
        """Initialize the listener model."""
        listener.attrs["_upstream_name"] = upstream_name
        listener.attrs[cls.SENTINEL] = True
        listener.attrs[cls.USE_DOC_ANNOTATIONS_FOR_PREDICTION] = True
        listener.attrs[cls.REQUIRES_ALL_LAYER_OUTPUTS] = requires_all_layer_outputs
        listener.attrs["_state"] = _ListenerNonPersistentState()

    @classmethod
    def use_doc_annotations_for_prediction(
        cls, listener: TransformerListenerModelT
    ) -> bool:
        """If True, the listener will perform its operations on the transformer output
        annotations stored on the Doc objects. Otherwise, it will perform its operations
        on the outputs that were stored directly in it using `TransformerListener.receive`."""
        return listener.attrs[cls.USE_DOC_ANNOTATIONS_FOR_PREDICTION]

    @classmethod
    def set_use_doc_annotations_for_prediction(
        cls, listener: TransformerListenerModelT, new_value: bool
    ):
        listener.attrs[cls.USE_DOC_ANNOTATIONS_FOR_PREDICTION] = new_value

    @classmethod
    def requires_all_layer_outputs(cls, model: Model) -> bool:
        return model.attrs[cls.REQUIRES_ALL_LAYER_OUTPUTS]

    @classmethod
    def get_upstream_name(cls, listener: TransformerListenerModelT) -> str:
        return listener.attrs["_upstream_name"]

    @staticmethod
    def get_batch_id(listener: TransformerListenerModelT) -> Optional[int]:
        state: _ListenerNonPersistentState = listener.attrs["_state"]
        return state.batch_id

    @staticmethod
    def get_output(
        listener: TransformerListenerModelT,
    ) -> Optional[TransformerModelOutput]:
        state: _ListenerNonPersistentState = listener.attrs["_state"]
        return state.outputs

    @staticmethod
    def get_backprop(
        listener: TransformerListenerModelT,
    ) -> Optional[Callable[[List[List[Ragged]], Tuple[int]], Any]]:
        state: _ListenerNonPersistentState = listener.attrs["_state"]
        return state.backprop

    @staticmethod
    def calculate_batch_id(inputs: Iterable[Doc]) -> int:
        """Calculate a content-sensitive hash of the batch of documents, to check
        whether the next batch of documents is unexpected.
        """
        return sum(sum(token.orth for token in doc) for doc in inputs)

    @staticmethod
    def receive(
        listener: TransformerListenerModelT,
        batch_id: int,
        outputs: TransformerModelOutput,
        backprop: Callable[[List[List[Ragged]], Tuple[int]], Any],
    ) -> None:
        """Store a batch of training predictions and a backprop callback in the listener.
        The predictions and callback are produced by the upstream Transformer component,
        and later will be used when the listener's component's model is called.
        """
        state: _ListenerNonPersistentState = listener.attrs["_state"]
        state.batch_id = batch_id
        state.outputs = outputs
        state.backprop = backprop

    @classmethod
    def verify_inputs(
        cls, listener: TransformerListenerModelT, inputs: Iterable[Doc]
    ) -> bool:
        """Check that the batch of Doc objects matches the ones we have a
        prediction for.
        """
        expected_batch_id = cls.get_batch_id(listener)
        outputs = cls.get_output(listener)

        if expected_batch_id is None and outputs is None:
            raise ValueError(SpacyErrors.E954)
        else:
            batch_id = cls.calculate_batch_id(inputs)
            if expected_batch_id != batch_id:
                raise ValueError(
                    SpacyErrors.E953.format(id1=expected_batch_id, id2=batch_id)
                )
            else:
                return True

    @staticmethod
    def clear_state(
        listener: TransformerListenerModelT,
    ):
        state: _ListenerNonPersistentState = listener.attrs["_state"]
        state.batch_id = None
        state.outputs = None
        state.backprop = None


def transformer_layers_listener_forward(
    model: TransformerListenerModelT, docs: Iterable[Doc], is_train: bool
) -> Tuple[List[List[Floats2d]], Callable[[Any], Any]]:
    pooling: WithRaggedLayersModelT = model.layers[0]
    grad_factor: float = model.attrs["grad_factor"]
    n_layers: int = model.attrs["layers"]

    _outputs = ListenerStateUtils.get_output(model)
    _backprop = ListenerStateUtils.get_backprop(model)

    if is_train:
        assert _outputs is not None
        if _outputs.last_layer_only:
            raise ValueError(
                Errors.E012.format(listener_name="TransformerLayersListener")
            )

        ListenerStateUtils.verify_inputs(model, docs)

        Y, backprop_pooling = pooling(_outputs.all_outputs, is_train)

        def backprop(dY):
            dX = backprop_pooling(dY)

            if grad_factor != 1.0:
                for dX_doc in dX:
                    for dX_layer in dX_doc:
                        dX_layer.data *= grad_factor

            outputs_to_backprop = tuple(i for i in range(0, _outputs.num_outputs))
            dX = _backprop(dX, outputs_to_backprop=outputs_to_backprop)

            ListenerStateUtils.clear_state(model)
            return dX

        return Y, backprop

    else:
        if ListenerStateUtils.use_doc_annotations_for_prediction(model):
            width = model.get_dim("nO")

            no_trf_data = [doc._.trf_data is None for doc in docs]
            if any(no_trf_data):
                assert all(no_trf_data)
                return [
                    [model.ops.alloc2f(len(doc), width) for _ in range(n_layers)]
                    for doc in docs
                ], lambda dY: []

            if any(doc._.trf_data.last_layer_only for doc in docs):
                raise ValueError(
                    Errors.E012.format(listener_name="TransformerLayersListener")
                )

            return pooling.predict(docs), lambda dY: []
        else:
            assert _outputs is not None
            ListenerStateUtils.verify_inputs(model, docs)
            outputs: Tuple[List[List[Floats2d]], Callable[[Any], Any]] = (
                pooling.predict(_outputs.all_outputs),
                lambda dY: [],
            )
            ListenerStateUtils.clear_state(model)
            return outputs


def last_transformer_layer_listener_forward(
    model: TransformerListenerModelT, docs: Iterable[Doc], is_train: bool
) -> Tuple[List[Floats2d], Callable[[Any], Any]]:
    pooling: WithRaggedLastLayerModelT = model.layers[0]
    grad_factor: float = model.attrs["grad_factor"]

    _outputs = ListenerStateUtils.get_output(model)
    _backprop = ListenerStateUtils.get_backprop(model)

    if is_train:
        assert _outputs is not None
        ListenerStateUtils.verify_inputs(model, docs)

        Y, backprop_pooling = pooling(_outputs.last_hidden_layer_states, is_train)

        def backprop(dY):
            dX_pooling = backprop_pooling(dY)
            if grad_factor != 1.0:
                for dx in dX_pooling:
                    dx.data *= grad_factor
            dX = _backprop([[d] for d in dX_pooling], outputs_to_backprop=(-1,))
            ListenerStateUtils.clear_state(model)

            return dX

        return Y, backprop
    else:
        if ListenerStateUtils.use_doc_annotations_for_prediction(model):
            width = model.get_dim("nO")

            no_trf_data = [doc._.trf_data is None for doc in docs]
            if any(no_trf_data):
                assert all(no_trf_data)
                return [
                    model.ops.alloc2f(len(doc), width) for doc in docs
                ], lambda dY: []

            return pooling.predict(docs), lambda dY: []
        else:
            assert _outputs is not None
            ListenerStateUtils.verify_inputs(model, docs)
            outputs: Tuple[List[Floats2d], Callable[[Any], Any]] = (
                pooling.predict(_outputs.last_hidden_layer_states),
                lambda dY: [],
            )
            ListenerStateUtils.clear_state(model)
            return outputs


def scalar_weighting_listener_forward(
    model: TransformerListenerModelT, docs: Iterable[Doc], is_train: bool
) -> Tuple[List[Floats2d], Callable[[Any], Any]]:
    weighting: ScalarWeightModelT = model.layers[0]
    pooling: WithRaggedLastLayerModelT = model.layers[1]
    grad_factor: float = model.attrs["grad_factor"]

    _outputs = ListenerStateUtils.get_output(model)
    _backprop = ListenerStateUtils.get_backprop(model)

    if is_train:
        assert _outputs is not None
        if _outputs.last_layer_only:
            raise ValueError(
                Errors.E012.format(listener_name="ScalarWeightingListener")
            )
        ListenerStateUtils.verify_inputs(model, docs)

        Y_weighting: ScalarWeightOutT = []
        weighting_inputs = _outputs.all_outputs
        outputs_to_backprop = tuple(i for i in range(_outputs.num_outputs))

        Y_weighting, backprop_weighting = weighting(weighting_inputs, is_train)
        Y, backprop_pooling = pooling(Y_weighting, is_train)

        def backprop(dYs):
            dX_pooling = backprop_pooling(dYs)
            dX_weighting = backprop_weighting(dX_pooling)

            if grad_factor != 1.0:
                for dx_inner in dX_weighting:
                    for dx in dx_inner:
                        dx.data *= grad_factor

            dX = _backprop(dX_weighting, outputs_to_backprop=outputs_to_backprop)
            ListenerStateUtils.clear_state(model)
            return dX

        return Y, backprop
    else:
        if ListenerStateUtils.use_doc_annotations_for_prediction(model):
            width = model.get_dim("nO")

            no_trf_data = [doc._.trf_data is None for doc in docs]
            if any(no_trf_data):
                assert all(no_trf_data)
                return [
                    model.ops.alloc2f(len(doc), width) for doc in docs
                ], lambda dY: []

            if any(doc._.trf_data.last_layer_only for doc in docs):
                raise ValueError(
                    Errors.E012.format(listener_name="ScalarWeightingListener")
                )

            Y_weighting = weighting.predict(
                [doc._.trf_data.all_outputs for doc in docs]
            )
            Y = pooling.predict(Y_weighting)

            return Y, lambda dX: []
        else:
            assert _outputs is not None
            ListenerStateUtils.verify_inputs(model, docs)
            Y_weighting = weighting.predict(_outputs.all_outputs)
            outputs: Tuple[List[Floats2d], Callable[[Any], Any]] = (
                pooling.predict(Y_weighting),
                lambda dY: [],
            )
            ListenerStateUtils.clear_state(model)
            return outputs


class WrappedTransformerAndListener(WrappedTransformerAndListenerModelT):
    """Wraps a transformer model and a compatible listener. Exclusively used when replacing
    listeners of a shared transformer pipeline."""

    name = "wrapped_transformer_and_listener"

    def __init__(
        self,
        transformer: TransformerModelT,
        listener: TransformerListenerModelT,
        frozen: bool = False,
    ) -> None:
        """
        transformer (TransformerModelT):
            Transformer model to wrap.
        listener (TransformerListenerModelT):
            Listener model to wrap.
        frozen (bool):
            If the transformer is frozen.
        """
        Model.__init__(
            self,
            name=self.name,
            init=wrapped_transformer_and_listener_init,
            forward=wrapped_transformer_and_listener_forward,
            dims={"nO": listener.get_dim("nO")},
            layers=[transformer, listener],
        )

        # Ensure that the transformer returns the required outputs.
        transformer.attrs[
            "_all_layer_outputs"
        ] = ListenerStateUtils.requires_all_layer_outputs(listener)

        # Freeze the embedded transformer if the source pipe was frozen.
        transformer.attrs["_frozen"] = frozen

        # Remove the marker from the wrapped listener so that it doesn't
        # get registered to any upstream transformer pipes.
        assert ListenerStateUtils.SENTINEL in listener.attrs
        del listener.attrs[ListenerStateUtils.SENTINEL]

        # Ensure that the listener directly reads from the last batch of transformer
        # outputs stored on it during prediction. If we don't do this, the wrapped
        # listener will end up using the annotations of the last upstream transformer
        # pipe that set the  `trf_data` annotation on the Doc object.
        assert ListenerStateUtils.use_doc_annotations_for_prediction(listener)
        ListenerStateUtils.set_use_doc_annotations_for_prediction(listener, False)

    @property
    def frozen_transformer(self) -> bool:
        return self.layers[0].attrs["_frozen"]

    @frozen_transformer.setter
    def frozen_transformer(self, value: bool):
        self.layers[0].attrs["_frozen"] = value


def wrapped_transformer_and_listener_init(model: WrappedTransformerAndListener, X, Y):
    transformer: TransformerModelT = model.layers[0]
    listener: TransformerListenerModelT = model.layers[1]

    transformer.init(X=X, Y=Y)
    listener.init(X=X, Y=Y)


def wrapped_transformer_and_listener_forward(
    model: WrappedTransformerAndListener,
    docs: WrappedTransformerAndListenerInT,
    is_train: bool,
) -> Tuple[WrappedTransformerAndListenerOutT, WrappedTransformerAndListenerBackpropT]:
    transformer: TransformerModelT = model.layers[0]
    listener: TransformerListenerModelT = model.layers[1]
    frozen: bool = transformer.attrs["_frozen"]
    ops = model.ops

    # Follows the same process as `Transformer.update()`.
    if frozen or not is_train:
        outputs = transformer.predict(docs)
        bp_outputs = None
        d_outputs = None
    else:
        outputs, bp_outputs = transformer(docs, is_train)
        d_outputs = [
            [Ragged(ops.alloc_f(t2v.dataXd.shape), t2v.lengths) for t2v in doc_layers]
            for doc_layers in outputs.all_outputs
        ]

    def backprop(
        one_d_outputs: List[List[Ragged]], outputs_to_backprop: Tuple[int, ...]
    ) -> Any:
        nonlocal d_outputs
        nonlocal frozen

        if frozen:
            return []

        assert bp_outputs is not None
        assert d_outputs is not None
        for i in range(len(one_d_outputs)):
            for j in outputs_to_backprop:
                d_outputs[i][j].data += one_d_outputs[i][j].data

        d_docs = bp_outputs(d_outputs)
        return d_docs

    # Set the output directly on the listener so that it can use them for both
    # training/backprop and prediction.
    ListenerStateUtils.receive(
        listener,
        batch_id=ListenerStateUtils.calculate_batch_id(docs),
        outputs=outputs,
        backprop=backprop,
    )
    listener_outputs, bp_listener_outputs = listener(docs, is_train)
    return listener_outputs, bp_listener_outputs


def replace_listener_callback(
    copied_trf_model: TransformerModelT,
    trf_listener: TransformerListenerModelT,
    trf_pipe: Any,
):
    # To avoid cyclic imports.
    from ..pipeline.transformer import CuratedTransformer

    assert isinstance(trf_pipe, CuratedTransformer)
    assert ListenerStateUtils.is_listener(trf_listener)

    copied_trf_listener = trf_listener.copy()
    wrapper = WrappedTransformerAndListener(
        transformer=copied_trf_model,
        listener=copied_trf_listener,
        frozen=trf_pipe.frozen,
    )
    return wrapper


def replace_listener_cfg_callback(trf_model_cfg, trf_listener_model_cfg):
    result = trf_model_cfg.copy()

    trf_model_cfg_arch = trf_model_cfg["@architectures"].split(".")
    assert len(trf_model_cfg_arch) == 3 and trf_model_cfg_arch[-2].endswith(
        "Transformer"
    )
    trf_listener_model_cfg_arch = trf_listener_model_cfg["@architectures"].split(".")
    assert len(trf_listener_model_cfg_arch) == 3 and trf_listener_model_cfg_arch[
        -2
    ].endswith("Listener")

    # The transformer model entrypoint should have a `wrapped_listener` parameter
    # that we can use to append the listener's config.
    result["wrapped_listener"] = trf_listener_model_cfg
    return result



from functools import partial
from typing import Any, Dict

import numpy
import pytest
import spacy
import torch
from spacy import Config, util
from spacy.language import Language
from spacy.training import Example
from spacy.training.initialize import init_nlp
from spacy.training.loop import train
from spacy.util import registry as spacy_registry
from spacy_curated_transformers._compat import has_hf_transformers, transformers
from spacy_curated_transformers.models.architectures import (
    build_bert_transformer_model_v1,
    build_camembert_transformer_model_v1,
    build_roberta_transformer_model_v1,
    build_xlmr_transformer_model_v1,
)
from spacy_curated_transformers.models.hf_loader import (
    build_hf_transformer_encoder_loader_v1,
)
from spacy_curated_transformers.models.listeners import (
    ListenerStateUtils,
    WrappedTransformerAndListener,
)
from spacy_curated_transformers.models.with_strided_spans import (
    build_with_strided_spans_v1,
)
from spacy_curated_transformers.pipeline.transformer import make_transformer
from spacy_curated_transformers.tokenization import (
    build_bert_wordpiece_encoder_v1,
    build_byte_bpe_encoder_v1,
    build_camembert_sentencepiece_encoder_v1,
    build_hf_piece_encoder_loader_v1,
    build_xlmr_sentencepiece_encoder_v1,
)
from spacy_curated_transformers.tokenization.sentencepiece_encoder import (
    build_sentencepiece_encoder_loader_v1,
)
from spacy_curated_transformers.util import create_gradual_transformer_unfreezing

from thinc.model import Model

from ..util import make_tempdir, torch_assertclose, xp_assert_array_equal

cfg_string_last_layer_listener = """
    # LastTransformerLayerListener

    [nlp]
    lang = "en"
    pipeline = ["transformer","tagger"]

    [components]

    [components.tagger]
    factory = "tagger"

    [components.tagger.model]
    @architectures = "spacy.Tagger.v2"
    nO = null

    [components.tagger.model.tok2vec]
    @architectures = "spacy-curated-transformers.LastTransformerLayerListener.v1"
    width = ${components.transformer.model.hidden_width}
    pooling = {"@layers":"reduce_mean.v1"}

    [components.transformer]
    factory = "curated_transformer"
    all_layer_outputs = False

    [components.transformer.model]
    @architectures = "spacy-curated-transformers.BertTransformer.v1"
    vocab_size = 28996
    num_hidden_layers = 1
    hidden_width = 60
    piece_encoder = {"@architectures":"spacy-curated-transformers.BertWordpieceEncoder.v1"}
    with_spans = {"@architectures":"spacy-curated-transformers.WithStridedSpans.v1"}

    [initialize]

    [initialize.components]

    [initialize.components.transformer]

    [initialize.components.transformer.piecer_loader]
    @model_loaders = "spacy-curated-transformers.HFPieceEncoderLoader.v1"
    name = "bert-base-cased"
"""

cfg_string_scalar_weighting_layer_listener = """
    # ScalarWeightingListener

    [nlp]
    lang = "en"
    pipeline = ["transformer","tagger"]

    [components]

    [components.tagger]
    factory = "tagger"

    [components.tagger.model]
    @architectures = "spacy.Tagger.v2"
    nO = null

    [components.tagger.model.tok2vec]
    @architectures = "spacy-curated-transformers.ScalarWeightingListener.v1"
    width = ${components.transformer.model.hidden_width}
    pooling = {"@layers":"reduce_mean.v1"}

    [components.tagger.model.tok2vec.weighting]
    @architectures = "spacy-curated-transformers.ScalarWeight.v1"
    num_layers = ${components.transformer.model.num_hidden_layers}

    [components.transformer]
    factory = "curated_transformer"
    all_layer_outputs = True

    [components.transformer.model]
    @architectures = "spacy-curated-transformers.BertTransformer.v1"
    vocab_size = 28996
    num_hidden_layers = 1
    hidden_width = 60
    piece_encoder = {"@architectures":"spacy-curated-transformers.BertWordpieceEncoder.v1"}
    with_spans = {"@architectures":"spacy-curated-transformers.WithStridedSpans.v1"}

    [initialize]

    [initialize.components]

    [initialize.components.transformer]

    [initialize.components.transformer.piecer_loader]
    @model_loaders = "spacy-curated-transformers.HFPieceEncoderLoader.v1"
    name = "bert-base-cased"
"""

TRAIN_DATA = [
    (
        "I like green eggs",
        {"tags": ["N", "V", "J", "N"], "cats": {"preference": 1.0, "imperative": 0.0}},
    ),
    (
        "Eat blue ham",
        {"tags": ["V", "J", "N"], "cats": {"preference": 0.0, "imperative": 1.0}},
    ),
]


def create_and_train_tagger(cfg_string):
    config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(config, auto_fill=True, validate=True)
    tagger = nlp.get_pipe("tagger")

    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
        for tag in t[1]["tags"]:
            tagger.add_label(tag)

    optimizer = nlp.initialize(lambda: train_examples)

    for _ in range(10):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    return nlp


def evaluate_tagger_on_train_data(model):
    docs = list(model.pipe(["Eat blue ham", "I like green eggs"]))
    assert [t.tag_ for t in docs[0]] == ["V", "J", "N"]
    assert [t.tag_ for t in docs[1]] == ["N", "V", "J", "N"]


def test_default_pipe_config_can_be_constructed():
    nlp = spacy.blank("en")
    nlp.add_pipe("curated_transformer", config={"model": {"vocab_size": 32}})


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize(
    "cfg_string",
    [cfg_string_last_layer_listener, cfg_string_scalar_weighting_layer_listener],
)
def test_tagger(cfg_string):
    model = create_and_train_tagger(cfg_string)
    evaluate_tagger_on_train_data(model)


def _hf_tokenize_per_token(tokenizer, docs, *, roberta=False):
    if roberta:
        hf_encoding = [
            tokenizer(
                [
                    doc[idx - 1].whitespace_ + token.text if idx > 0 else token.text
                    for idx, token in enumerate(doc)
                ]
            )
            for doc in docs
        ]
    else:
        hf_encoding = [tokenizer([token.text for token in doc]) for doc in docs]
    ids = []
    lens = []
    bos_id = (
        tokenizer.bos_token_id
        if tokenizer.bos_token_id is not None
        else tokenizer.cls_token_id
    )
    eos_id = (
        tokenizer.eos_token_id
        if tokenizer.eos_token_id is not None
        else tokenizer.sep_token_id
    )
    for i in range(len(hf_encoding)):
        doc_ids = [id for e in hf_encoding[i]["input_ids"] for id in e[1:-1]]
        ids.append([bos_id] + doc_ids + [eos_id])
        lens.append(len(ids[-1]))

    torch_ids = torch.full(
        (len(ids), max(lens)), tokenizer.pad_token_id, dtype=torch.int
    )
    for i in range(len(ids)):
        torch_ids[i][: len(ids[i])] = torch.tensor(ids[i])

    attention_mask = torch_ids.ne(tokenizer.pad_token_id)

    return torch_ids, attention_mask, lens


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_bert_transformer_pipe_against_hf():
    nlp = spacy.blank("en")
    model = build_bert_transformer_model_v1(
        piece_encoder=build_bert_wordpiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=28996,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="bert-base-cased"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="bert-base-cased"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("bert-base-cased")
    hf_model = transformers.AutoModel.from_pretrained("bert-base-cased")

    docs = [
        nlp.make_doc("I saw a girl with a telescope."),
        nlp.make_doc("Today we will eat poké bowl."),
    ]

    hf_ids, attention_mask, lens = _hf_tokenize_per_token(hf_tokenizer, docs)
    hf_encoding = hf_model(hf_ids, attention_mask=attention_mask)
    docs = list(pipe.pipe(docs))

    for doc, hf_doc_encoding, encoding_len in zip(
        docs, hf_encoding.last_hidden_state, lens
    ):
        torch_assertclose(
            hf_doc_encoding[:encoding_len][1:-1],
            torch.tensor(doc._.trf_data.last_hidden_layer_state.dataXd),
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_camembert_transformer_pipe_against_hf():
    nlp = spacy.blank("fr")
    model = build_camembert_transformer_model_v1(
        piece_encoder=build_camembert_sentencepiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=32005,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="camembert-base"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="camembert-base"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("camembert-base")
    hf_model = transformers.AutoModel.from_pretrained("camembert-base")

    docs = [
        nlp.make_doc("J'ai vu une fille avec un télescope."),
        nlp.make_doc("Aujourd'hui, nous allons manger un poké bowl."),
    ]

    hf_ids, attention_mask, lens = _hf_tokenize_per_token(hf_tokenizer, docs)
    hf_encoding = hf_model(hf_ids, attention_mask=attention_mask)
    docs = list(pipe.pipe(docs))

    for doc, hf_doc_encoding, encoding_len in zip(
        docs, hf_encoding.last_hidden_state, lens
    ):
        torch_assertclose(
            hf_doc_encoding[:encoding_len][1:-1],
            torch.tensor(doc._.trf_data.last_hidden_layer_state.dataXd),
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_roberta_transformer_pipe_against_hf():
    nlp = spacy.blank("en")
    model = build_roberta_transformer_model_v1(
        piece_encoder=build_byte_bpe_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=50265,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="roberta-base"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="roberta-base"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("roberta-base")
    hf_model = transformers.AutoModel.from_pretrained("roberta-base")

    docs = [
        nlp.make_doc("I saw a girl with a telescope."),
        nlp.make_doc("Today we will eat poké bowl."),
    ]

    hf_ids, attention_mask, lens = _hf_tokenize_per_token(
        hf_tokenizer, docs, roberta=True
    )
    hf_encoding = hf_model(hf_ids, attention_mask=attention_mask)
    docs = list(pipe.pipe(docs))

    for doc, hf_doc_encoding, encoding_len in zip(
        docs, hf_encoding.last_hidden_state, lens
    ):
        torch_assertclose(
            hf_doc_encoding[:encoding_len][1:-1],
            torch.tensor(doc._.trf_data.last_hidden_layer_state.dataXd),
        )


def test_empty_input(sentencepiece_toy_model_path):
    nlp = spacy.blank("en")
    model = build_xlmr_transformer_model_v1(
        piece_encoder=build_xlmr_sentencepiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        num_hidden_layers=2,
        vocab_size=1000,
        hidden_width=12,
    )

    model.get_ref("piece_encoder").init = build_sentencepiece_encoder_loader_v1(
        path=sentencepiece_toy_model_path
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)
    pipe(nlp.make_doc(""))


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_xlmr_transformer_pipe_against_hf():
    nlp = spacy.blank("en")
    model = build_xlmr_transformer_model_v1(
        piece_encoder=build_xlmr_sentencepiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=250002,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("xlm-roberta-base")
    hf_model = transformers.AutoModel.from_pretrained("xlm-roberta-base")

    docs = [
        nlp.make_doc("I saw a girl with a telescope."),
        nlp.make_doc("Today we will eat poké bowl."),
    ]

    hf_ids, attention_mask, lens = _hf_tokenize_per_token(hf_tokenizer, docs)
    hf_encoding = hf_model(hf_ids, attention_mask=attention_mask)
    docs = list(pipe.pipe(docs))

    for doc, hf_doc_encoding, encoding_len in zip(
        docs, hf_encoding.last_hidden_state, lens
    ):
        torch_assertclose(
            hf_doc_encoding[:encoding_len][1:-1],
            torch.tensor(doc._.trf_data.last_hidden_layer_state.dataXd),
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_frozen_transformer_pipe():
    config = Config().from_str(cfg_string_scalar_weighting_layer_listener)
    nlp = util.load_model_from_config(config, auto_fill=True, validate=True)
    tagger = nlp.get_pipe("tagger")
    transformer = nlp.get_pipe("transformer")
    transformer.frozen = True

    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
        for tag in t[1]["tags"]:
            tagger.add_label(tag)

    optimizer = nlp.initialize(lambda: train_examples)

    def get_transformer_params_sorted():
        params = transformer.model.get_ref("transformer").shims[0]._model.state_dict()
        return list(sorted(params.items()))

    transformer_init_params = [
        (k, v.clone()) for k, v in get_transformer_params_sorted()
    ]

    for i in range(5):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    transformer_trained_params = get_transformer_params_sorted()
    for (old_param, old_vec), (new_param, new_vec) in zip(
        transformer_init_params, transformer_trained_params
    ):
        assert old_param == new_param
        torch_assertclose(
            old_vec,
            new_vec,
        )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_transformer_pipe_outputs():
    nlp = spacy.blank("en")
    model = build_xlmr_transformer_model_v1(
        piece_encoder=build_xlmr_sentencepiece_encoder_v1(),
        with_spans=build_with_strided_spans_v1(),
        vocab_size=250002,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.initialize()
    pipe = make_transformer(nlp, "transformer", model, all_layer_outputs=False)

    docs = [
        nlp.make_doc("I saw a girl with a telescope."),
        nlp.make_doc("Today we will eat poké bowl."),
    ]
    docs = list(pipe.pipe(docs))
    assert all([doc._.trf_data.last_layer_only for doc in docs]) == True
    assert all([len(doc._.trf_data.all_outputs) == 1 for doc in docs]) == True

    pipe = make_transformer(nlp, "transformer", model, all_layer_outputs=True)
    docs = list(pipe.pipe(docs))
    assert all([not doc._.trf_data.last_layer_only for doc in docs]) == True
    assert all([len(doc._.trf_data.all_outputs) == 12 + 1 for doc in docs]) == True


cfg_string_gradual_unfreezing = (
    cfg_string_last_layer_listener
    + """
    [corpora]

    [corpora.train]
    @readers = "spacy.Corpus.v1"
    path = "toy-en-corpus.spacy"
    max_length = 500
    gold_preproc = false
    limit = 0

    [corpora.dev]
    @readers = "spacy.Corpus.v1"
    path = "toy-en-corpus.spacy"
    max_length = 500
    gold_preproc = false
    limit = 0

    [training]
    train_corpus = "corpora.train"
    dev_corpus = "corpora.dev"
    seed = 1
    gpu_allocator = "pytorch"
    dropout = 0.1
    accumulate_gradient = 3
    patience = 5000
    max_epochs = 1
    max_steps = 6
    eval_frequency = 10

    [training.batcher]
    @batchers = "spacy.batch_by_padded.v1"
    discard_oversize = False
    get_length = null
    size = 1
    buffer = 256
"""
)


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_gradual_transformer_unfreezing(test_dir):
    def wrapped_callback():
        def inner(
            nlp: Language,
            args: Dict[str, Any],
            gradual_unfreezing_callback,
            unfreeze_step,
        ):
            current_step = args["step"]
            gradual_unfreezing_callback(nlp, args)

            transformer = nlp.get_pipe("transformer")
            if current_step < unfreeze_step:
                assert transformer.frozen == True
            else:
                assert transformer.frozen == False

        return partial(
            inner,
            gradual_unfreezing_callback=create_gradual_transformer_unfreezing({"*": 3}),
            unfreeze_step=3,
        )

    spacy_registry.callbacks.register(
        "test_gradual_unfreezing_callback",
        func=wrapped_callback,
    )

    config = Config().from_str(cfg_string_gradual_unfreezing, interpolate=False)
    config["corpora"]["train"]["path"] = str(test_dir / "toy-en-corpus.spacy")
    config["corpora"]["dev"]["path"] = str(test_dir / "toy-en-corpus.spacy")
    config["training"]["before_update"] = {
        "@callbacks": "test_gradual_unfreezing_callback"
    }
    nlp = util.load_model_from_config(config, auto_fill=True, validate=True)

    nlp = init_nlp(config)
    train(nlp)
    assert nlp.get_pipe("transformer").frozen == False

    with pytest.raises(ValueError):
        create_gradual_transformer_unfreezing({"transformer": 5, "*": 4})


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize(
    ["cfg_string", "listener_name", "listener_entrypoint"],
    [
        (
            cfg_string_last_layer_listener,
            "last_transformer_layer_listener",
            "spacy-curated-transformers.LastTransformerLayerListener.v1",
        ),
        (
            cfg_string_scalar_weighting_layer_listener,
            "scalar_weighting_listener",
            "spacy-curated-transformers.ScalarWeightingListener.v1",
        ),
    ],
)
def test_replace_listeners(cfg_string, listener_name, listener_entrypoint):
    orig_config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    text = "This is awesome"
    examples = [Example.from_dict(nlp.make_doc(text), {"tags": ["A", "B", "C"]})]
    optimizer = nlp.initialize(lambda: examples)

    # verify correct configuration with transformer listener
    transformer = nlp.get_pipe("transformer")
    tagger = nlp.get_pipe("tagger")

    tagger_tok2vec = tagger.model.get_ref("tok2vec")

    assert ListenerStateUtils.is_listener(tagger_tok2vec)
    assert transformer.listener_map["tagger"][0] == tagger_tok2vec
    assert isinstance(transformer.model, Model)
    assert transformer.model.name == "transformer_model"
    assert (
        nlp.config["components"]["transformer"]["model"]["@architectures"]
        == "spacy-curated-transformers.BertTransformer.v1"
    )
    assert (
        nlp.config["components"]["tagger"]["model"]["tok2vec"]["@architectures"]
        == listener_entrypoint
    )

    # train pipe before replacing listeners
    for i in range(5):
        losses = {}
        nlp.update(examples, sgd=optimizer, losses=losses)
        doc = nlp(text)

    preds = [t.tag_ for t in doc]
    doc_tensor = tagger_tok2vec.predict([doc])

    # replace listener and verify predictions are still the same
    transformer.frozen = True
    nlp.replace_listeners("transformer", "tagger", ["model.tok2vec"])
    tagger = nlp.get_pipe("tagger")
    tagger_tok2vec = tagger.model.get_ref("tok2vec")
    assert isinstance(tagger_tok2vec, WrappedTransformerAndListener)
    assert tagger_tok2vec.frozen_transformer
    assert tagger_tok2vec.layers[0].name == "transformer_model"
    assert tagger_tok2vec.layers[1].name == listener_name
    assert (
        nlp.config["components"]["tagger"]["model"]["tok2vec"]["@architectures"]
        == "spacy-curated-transformers.BertTransformer.v1"
    )
    assert (
        nlp.config["components"]["tagger"]["model"]["tok2vec"]["wrapped_listener"][
            "@architectures"
        ]
        == listener_entrypoint
    )
    doc2 = nlp(text)
    assert preds == [t.tag_ for t in doc2]
    pred_tensor = tagger_tok2vec.predict([doc2])
    xp_assert_array_equal(doc_tensor, pred_tensor)

    optimizer = nlp.resume_training()
    trf_output_frozen = tagger_tok2vec.layers[0].predict([doc2])
    for i in range(5):
        losses = {}
        nlp.update(examples, sgd=optimizer, losses=losses)
        assert losses["tagger"] > 0.0
    trf_output_frozen_after_update = tagger_tok2vec.layers[0].predict([doc2])

    for x, y in zip(
        trf_output_frozen.all_outputs, trf_output_frozen_after_update.all_outputs
    ):
        for x1, y1 in zip(x, y):
            xp_assert_array_equal(x1.dataXd, y1.dataXd)

    tagger_tok2vec.frozen_transformer = False
    trf_output = tagger_tok2vec.layers[0].predict([doc2])
    # attempt training with the new pipeline
    for i in range(5):
        losses = {}
        nlp.update(examples, sgd=optimizer, losses=losses)
        assert losses["tagger"] > 0.0
    trained_trf_output = tagger_tok2vec.layers[0].predict([doc2])

    for x, y in zip(trf_output.all_outputs, trained_trf_output.all_outputs):
        for x1, y1 in zip(x, y):
            assert not numpy.array_equal(x1.dataXd, y1.dataXd)

    # ensure IO goes OK
    doc_tensor_trained = tagger_tok2vec.predict([doc])

    with make_tempdir() as d:
        file_path = d / "trained_nlp"
        nlp.to_disk(file_path)
        nlp2 = util.load_model_from_path(file_path)
        doc3 = nlp2(text)
        tagger2 = nlp2.get_pipe("tagger")
        tagger_tok2vec2 = tagger2.model.get_ref("tok2vec")
        pred_tensor = tagger_tok2vec2.predict([doc3])
        xp_assert_array_equal(doc_tensor_trained, pred_tensor)



import pytest
from spacy.util import registry as spacy_registry
from spacy_curated_transformers.util import registry


@pytest.mark.parametrize(
    "model_name",
    [
        "spacy-curated-transformers.AlbertTransformer.v1",
        "spacy-curated-transformers.BertTransformer.v1",
        "spacy-curated-transformers.CamembertTransformer.v1",
        "spacy-curated-transformers.RobertaTransformer.v1",
        "spacy-curated-transformers.XlmrTransformer.v1",
        "spacy-curated-transformers.WithStridedSpans.v1",
        "spacy-curated-transformers.ScalarWeight.v1",
        "spacy-curated-transformers.TransformerLayersListener.v1",
        "spacy-curated-transformers.LastTransformerLayerListener.v1",
        "spacy-curated-transformers.ScalarWeightingListener.v1",
    ],
)
def test_model_from_registry(model_name):
    # Can't be constructed, since all models have mandatory arguments.
    spacy_registry.architectures.get(model_name)


@pytest.mark.parametrize(
    "encoder_name",
    [
        "spacy-curated-transformers.BertWordpieceEncoder.v1",
        "spacy-curated-transformers.ByteBpeEncoder.v1",
        "spacy-curated-transformers.CamembertSentencepieceEncoder.v1",
        "spacy-curated-transformers.CharEncoder.v1",
        "spacy-curated-transformers.SentencepieceEncoder.v1",
        "spacy-curated-transformers.WordpieceEncoder.v1",
        "spacy-curated-transformers.XlmrSentencepieceEncoder.v1",
    ],
)
def test_tokenizer_encoder_from_registry(encoder_name):
    spacy_registry.architectures.get(encoder_name)()


@pytest.mark.parametrize(
    "loader_name",
    [
        "spacy-curated-transformers.ByteBpeLoader.v1",
        "spacy-curated-transformers.CharEncoderLoader.v1",
        "spacy-curated-transformers.HFTransformerEncoderLoader.v1",
        "spacy-curated-transformers.HFPieceEncoderLoader.v1",
        "spacy-curated-transformers.PyTorchCheckpointLoader.v1",
        "spacy-curated-transformers.SentencepieceLoader.v1",
        "spacy-curated-transformers.WordpieceLoader.v1",
    ],
)
def test_model_loaders_from_registry(loader_name):
    # Can't be constructed, since most loaders have mandatory arguments.
    registry.model_loaders.get(loader_name)


@pytest.mark.parametrize(
    "callback_name",
    [
        "spacy-curated-transformers.gradual_transformer_unfreezing.v1",
    ],
)
def test_callbacks_from_registry(callback_name):
    # Can't be constructed, since all callbacks have mandatory arguments.
    spacy_registry.callbacks.get(callback_name)



class ErrorsWithCodes(type):
    def __getattribute__(self, code):
        msg = super().__getattribute__(code)
        if code.startswith("__"):  # python system attributes like __class__
            return msg
        else:
            return "(spacy-curated-transformers) [{code}] {msg}".format(
                code=code, msg=msg
            )


# fmt: off

class Errors(metaclass=ErrorsWithCodes):
    E009 = ("At least one sequence in the transformer's input has a length "
            "of {seq_len}, which is larger than the model's maximum sequence "
            "length of {max_seq_len} tokens")
    E010 = ("Curated transformers do not currently support listener replacement")
    E011 = ("`{loader_name}` requires the Hugging Face `transformers` package to be installed")
    E012 = ("`{listener_name}` requires the upstream transformer pipe to output "
            "all hidden layer outputs. This can be enabled by setting the pipe's "
            "`all_layer_outputs` parameter to `True` in the pipeline config")
    E013 = ("The target pipe names for gradual transformer unfreezing contain both the "
            "wild-card operator ('*') and individual names. Use either of the two but not both")
    E014 = ("Model '{model_name}' received an unexpected input of type '{input_type}'. "
            "It can only wrap/be chained with models whose outputs are of type  "
            "`TransformerModelOutput` (in almost all cases, models of type `TorchTransformerModelT`)")
    E015 = ("Input passed to the `ScalarWeight` model do not have the same number "
            "layers. Distinct layer counts: {layer_counts}")
    E016 = ("Input passed to the `ScalarWeight` model do not have the same width. "
            "Distinct widths: {hidden_widths}")
    E017 = ("Span extractor stride ({stride}) must be within [window_size / 2, "
            "window_size] ([{half_window_size}, {window_size}])")
    E018 = ("Span extractor batch size must be greater than zero")
    E019 = ("Byte-BPE piece encoder vocabulary doesn't contain '{piece}' piece")
    E020 = ("Character piece encoder vocabulary is not available. Use a loader "
            "to initialize the encoder.")
    E021 = ("Attempting to use the `CharEncoderLoader` piece encoder loader with an "
            "incompatible model ('{model_name}'). It can only be used with the "
            "`CharEncoder` piece encoder")
    E022 = ("Attempting to load an unsupported Hugging Face tokenizer "
            "({unsupported_tokenizer}). Currently supported tokenizers: "
            "{supported_tokenizers}")
    E023 = ("Japanese BERT models currently only support character subword encoding")
    E024 = ("Attempting to initialize an incompatible piece encoder ('{model_name}') "
            "with the Hugging Face Japanese BERT tokenizer. It can only be used with the "
            "`CharEncoder` piece encoder")
    E025 = ("Attempting to perform gradual unfreezing of a non-transformer pipe "
            "('{pipe_name}'}. Only transformer pipes support this feature")
    E026 = ("Attempting to register a model ('{model_name}') with the transformer pipe"
            "that isn't a transformer listener")

# fmt: on



import contextlib
import shutil
import tempfile
from pathlib import Path

import torch

from thinc.api import NumpyOps


@contextlib.contextmanager
def make_tempdir():
    d = Path(tempfile.mkdtemp())
    yield d
    shutil.rmtree(str(d))


# Wrapper around torch.testing.assert_close with custom tolerances.
def torch_assertclose(
    a: torch.Tensor, b: torch.Tensor, *, atol: float = 1e-05, rtol: float = 1e-05
):
    torch.testing.assert_close(
        a,
        b,
        atol=atol,
        rtol=rtol,
    )


def xp_assert_array_equal(a, b):
    # Always convert to CPU since Cupy will strangely complain
    # about requiring a Numpy conversion before calling assert_array_close.
    if not isinstance(a, list):
        a = [a]
    if not isinstance(b, list):
        b = [b]
    ops = NumpyOps()
    for x, y in zip(a, b):
        ops.xp.testing.assert_array_equal(ops.asarray(x), ops.asarray(y))
